@@ -67,6 +67,8 @@ pub struct Translate {
     pub universal: bool,
     #[serde(default, rename = "user-db")]
     pub userdb: bool,
+    #[serde(default, rename = "drop-raw")]
+    pub drop_raw: bool,
 }
 
 fn execve_env_default() -> HashSet<String> {
@@ -86,7 +88,7 @@ fn false_value() -> bool {
 
 #[derive(Debug, Serialize, Deserialize)]
 pub struct Enrich {
-    #[serde(default, rename = "execve-env")]
+    #[serde(default = "execve_env_default", rename = "execve-env")]
     pub execve_env: HashSet<String>,
     #[serde(default = "true_value")]
     pub container: bool,
@@ -126,14 +128,47 @@ pub struct LabelProcess {
     pub propagate_labels: HashSet<String>,
 }
 
+#[derive(Default, Debug, Serialize, Deserialize)]
+#[serde(rename_all = "lowercase")]
+pub enum FilterAction {
+    #[default]
+    #[serde(alias = "Drop")]
+    Drop,
+    #[serde(alias = "Log")]
+    Log,
+}
+
+pub(crate) mod regex_set {
+    use regex::bytes::RegexSet;
+    use serde::{Deserialize, Deserializer, Serialize, Serializer};
+
+    pub(crate) fn serialize<S>(v: &RegexSet, s: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        v.patterns().to_vec().serialize(s)
+    }
+    pub(crate) fn deserialize<'de, D>(d: D) -> Result<RegexSet, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        let v: Vec<String> = Deserialize::deserialize(d)?;
+        RegexSet::new(v).map_err(serde::de::Error::custom)
+    }
+}
+
 #[derive(Default, Debug, Serialize, Deserialize)]
 pub struct Filter {
     #[serde(default, rename = "filter-keys")]
     pub filter_keys: HashSet<String>,
     #[serde(default, rename = "filter-labels")]
     pub filter_labels: HashSet<String>,
+    #[serde(default, rename = "filter-raw-lines", with = "regex_set")]
+    pub filter_raw_lines: regex::bytes::RegexSet,
     #[serde(default, rename = "filter-null-keys")]
     pub filter_null_keys: bool,
+    #[serde(default, rename = "filter-action")]
+    pub filter_action: FilterAction,
 }
 
 #[derive(Debug, Serialize)]
@@ -198,6 +233,8 @@ pub struct Config {
     #[serde(default)]
     pub auditlog: Logfile,
     #[serde(default)]
+    pub filterlog: Logfile,
+    #[serde(default)]
     pub debug: Debug,
     #[serde(default)]
     pub transform: Transform,
@@ -225,6 +262,13 @@ impl Default for Config {
                 generations: Some(5),
                 line_prefix: None,
             },
+            filterlog: Logfile {
+                file: "filtered.log".into(),
+                users: None,
+                size: Some(10 * 1024 * 1024),
+                generations: Some(5),
+                line_prefix: None,
+            },
             debug: Debug::default(),
             transform: Transform::default(),
             translate: Translate::default(),
@@ -288,10 +332,11 @@ impl Config {
                 .collect(),
             translate_universal: self.translate.universal,
             translate_userdb: self.translate.userdb,
-            label_exe: self.label_process.label_exe.as_ref(),
-            unlabel_exe: self.label_process.unlabel_exe.as_ref(),
-            label_script: self.label_process.label_script.as_ref(),
-            unlabel_script: self.label_process.unlabel_script.as_ref(),
+            drop_translated: self.translate.drop_raw,
+            label_exe: self.label_process.label_exe.clone(),
+            unlabel_exe: self.label_process.unlabel_exe.clone(),
+            label_script: self.label_process.label_script.clone(),
+            unlabel_script: self.label_process.unlabel_script.clone(),
             filter_keys: self
                 .filter
                 .filter_keys
@@ -305,6 +350,7 @@ impl Config {
                 .map(|s| s.as_bytes().to_vec())
                 .collect(),
             filter_null_keys: self.filter.filter_null_keys,
+            filter_raw_lines: self.filter.filter_raw_lines.clone(),
         }
     }
 }
@@ -348,11 +394,15 @@ read-users = ["splunk"]
 
     #[test]
     fn parse_defaults() {
-        toml::de::from_str::<Config>("").unwrap();
+        let cfg_default = toml::de::from_str::<Config>("").unwrap();
+        println!("{}", toml::to_string(&cfg_default).unwrap());
 
-        toml::de::from_str::<Config>(
+        println!("--------------------");
+
+        let cfg_empty_sections = toml::de::from_str::<Config>(
             r#"
 [auditlog]
+[filterlog]
 [debug]
 [transform]
 [translate]
@@ -362,5 +412,9 @@ read-users = ["splunk"]
 "#,
         )
         .unwrap();
+        println!("{}", toml::to_string(&cfg_empty_sections).unwrap());
+
+        // FIXME This does not work because HashSet ordering is not stable.
+        // assert!(toml::to_string(&cfg_default) == toml::to_string(&cfg_empty_sections));
     }
 }
