@@ -1,6 +1,6 @@
 use std::convert::{From, TryFrom};
 use std::ops::Range;
-use std::str::{self, FromStr};
+use std::str;
 
 use crate::constants::*;
 use crate::types::*;
@@ -10,6 +10,26 @@ use nom::{
     sequence::*, IResult, Offset,
 };
 
+use nom::character::complete::{i64 as dec_i64, u16 as dec_u16, u32 as dec_u32, u64 as dec_u64};
+
+use thiserror::Error;
+
+#[derive(Debug, Error)]
+pub enum ParseError {
+    #[error("cannot parse header: {}", String::from_utf8_lossy(.0))]
+    MalformedHeader(Vec<u8>),
+    #[error("cannot parse body: {}", String::from_utf8_lossy(.0))]
+    MalformedBody(Vec<u8>),
+    #[error("garbage at end of message: {}", String::from_utf8_lossy(.0))]
+    TrailingGarbage(Vec<u8>),
+    #[error("{id} ({ty}) can't hex-decode {}", String::from_utf8_lossy(.hex_str))]
+    HexDecodeError {
+        ty: MessageType,
+        id: EventID,
+        hex_str: Vec<u8>,
+    },
+}
+
 /// Parse a single log line as produced by _auditd(8)_
 ///
 /// If `skip_enriched` is set and _auditd_ has been configured to
@@ -20,26 +40,15 @@ use nom::{
 pub fn parse(
     mut raw: Vec<u8>,
     skip_enriched: bool,
-) -> Result<(Option<Vec<u8>>, MessageType, EventID, Record), String> {
-    let (rest, (nd, ty, id)) = parse_header(&raw).map_err(|e| {
-        format!(
-            "cannot parse header: {}",
-            e.map_input(String::from_utf8_lossy)
-        )
-    })?;
+) -> Result<(Option<Vec<u8>>, MessageType, EventID, Record), ParseError> {
+    let (rest, (nd, ty, id)) =
+        parse_header(&raw).map_err(|_| ParseError::MalformedHeader(raw.clone()))?;
 
-    let (rest, body) = parse_body(rest, ty, skip_enriched).map_err(|e| {
-        format!(
-            "cannot parse body: {}",
-            e.map_input(String::from_utf8_lossy)
-        )
-    })?;
+    let (rest, body) = parse_body(rest, ty, skip_enriched)
+        .map_err(|_| ParseError::MalformedBody(rest.to_vec()))?;
 
     if !rest.is_empty() {
-        return Err(format!(
-            "garbage at end of message: {}",
-            String::from_utf8_lossy(rest)
-        ));
+        return Err(ParseError::TrailingGarbage(rest.to_vec()));
     }
 
     let nd = nd.map(|s| s.to_vec());
@@ -80,10 +89,12 @@ pub fn parse(
             let d = unsafe {
                 str::from_utf8_unchecked(&raw[stride.start + 2 * i..stride.start + 2 * i + 2])
             };
-            raw[stride.start + i] = u8::from_str_radix(d, 16).map_err(|_| {
-                let hex_str = unsafe { str::from_utf8_unchecked(&raw[stride.clone()]) };
-                format!("{} ({}) can't hex-decode {}", id, ty, hex_str)
-            })?;
+            raw[stride.start + i] =
+                u8::from_str_radix(d, 16).map_err(|_| ParseError::HexDecodeError {
+                    id,
+                    ty,
+                    hex_str: raw[stride.clone()].to_vec(),
+                })?;
         }
     }
 
@@ -128,10 +139,7 @@ fn parse_type(input: &[u8]) -> IResult<&[u8], MessageType> {
                         .map(|n| MessageType(*n))
                 },
             ),
-            map_res(
-                delimited(tag("UNKNOWN["), is_a("0123456789"), tag("]")),
-                |s| str::from_utf8(s).unwrap().parse::<u32>().map(MessageType),
-            ),
+            map(delimited(tag("UNKNOWN["), dec_u32, tag("]")), MessageType),
         )),
     )(input)
 }
@@ -139,28 +147,20 @@ fn parse_type(input: &[u8]) -> IResult<&[u8], MessageType> {
 /// Recognize the "msg=audit(…):" event identifier
 #[inline(always)]
 fn parse_msgid(input: &[u8]) -> IResult<&[u8], EventID> {
-    map_res(
+    map(
         tuple((
-            tag("msg=audit("),
-            digit1,
-            tag("."),
-            digit1,
-            tag(":"),
-            digit1,
-            tag("):"),
-            take_while(is_space),
+            preceded(tag("msg=audit("), dec_u64),
+            delimited(tag("."), dec_u64, tag(":")),
+            terminated(dec_u32, pair(tag("):"), space0)),
         )),
-        |(_, sec, _, msec, _, seq, _, _)| -> Result<EventID, std::num::ParseIntError> {
-            Ok(EventID {
-                // safety: captured strings contain only ASCII digits
-                timestamp: 1000 * unsafe { str::from_utf8_unchecked(sec) }.parse::<u64>()?
-                    + unsafe { str::from_utf8_unchecked(msec) }.parse::<u64>()?,
-                sequence: unsafe { str::from_utf8_unchecked(seq) }.parse::<u32>()?,
-            })
+        |(sec, msec, sequence)| EventID {
+            timestamp: 1000 * sec + msec,
+            sequence,
         },
     )(input)
 }
 
+#[derive(Clone)]
 enum PValue<'a> {
     Empty,
     HexStr(&'a [u8]),
@@ -177,30 +177,35 @@ fn parse_body(
     ty: MessageType,
     skip_enriched: bool,
 ) -> IResult<&[u8], Vec<(Key, PValue)>> {
-    let (input, special) = opt(alt((
-        map_res(
+    // Handle some corner cases that don't fit the general key=value
+    // scheme.
+    let (input, special) = match ty {
+        msg_type::AVC => opt(map(
             tuple((
-                tuple((tag("avc:"), space0)),
-                alt((tag("granted"), tag("denied"))),
-                tuple((space0, tag("{"), space0)),
-                many1(terminated(parse_identifier, space0)),
-                tuple((tag("}"), space0, tag("for"), space0)),
+                preceded(
+                    pair(tag("avc:"), space0),
+                    alt((tag("granted"), tag("denied"))),
+                ),
+                delimited(
+                    tuple((space0, tag("{"), space0)),
+                    many1(terminated(parse_identifier, space0)),
+                    tuple((tag("}"), space0, tag("for"), space0)),
+                ),
             )),
-            |(_, k, _, v, _)| -> Result<_, ()> { Ok((Key::Name(NVec::from(k)), PValue::List(v))) },
-        ),
-        map_res(
-            tuple((tag("netlabel"), tag(":"), space0)),
-            |(s, _, _): (&[u8], _, _)| -> Result<_, ()> {
-                Ok((Key::Name(NVec::from(s)), PValue::Empty))
-            },
-        ),
-    )))(input)?;
-
-    let (input, _) = match ty {
-        // Skip. overe start of message doesn't fit the key=value
-        // scheme and does not contain useful information.
-        msg_type::MAC_POLICY_LOAD => opt(tag("policy loaded "))(input)?,
-        _ => (input, None),
+            |(k, v)| (Key::Name(NVec::from(k)), PValue::List(v)),
+        ))(input)?,
+        msg_type::TTY => {
+            let (input, _) = opt(tag("tty "))(input)?;
+            (input, None)
+        }
+        msg_type::MAC_POLICY_LOAD => {
+            let (input, _) = opt(tag("policy loaded "))(input)?;
+            (input, None)
+        }
+        _ => opt(map(
+            terminated(tag("netlabel"), pair(tag(":"), space0)),
+            |s| (Key::Name(NVec::from(s)), PValue::Empty),
+        ))(input)?,
     };
 
     let (input, mut kv) = if skip_enriched {
@@ -232,26 +237,30 @@ fn parse_body(
 fn parse_kv(input: &[u8], ty: MessageType) -> IResult<&[u8], (Key, PValue)> {
     let (input, key) = match ty {
         // Special case for execve arguments: aX, aX[Y], aX_len
-        msg_type::EXECVE if !input.is_empty() && input[0] == b'a' => terminated(
-            alt((parse_key_a_x_len, parse_key_a_xy, parse_key_a_x, parse_key)),
-            tag("="),
-        )(input),
+        msg_type::EXECVE
+            if !input.is_empty() && input[0] == b'a' && !input.starts_with(b"argc") =>
+        {
+            terminated(
+                alt((parse_key_a_x_len, parse_key_a_xy, parse_key_a_x)),
+                tag("="),
+            )(input)
+        }
         // Special case for syscall params: aX
         msg_type::SYSCALL => terminated(alt((parse_key_a_x, parse_key)), tag("="))(input),
         _ => terminated(parse_key, tag("="))(input),
     }?;
 
     let (input, value) = match (ty, &key) {
-        (msg_type::SYSCALL, Key::Arg(_, None)) => map_res(
+        (msg_type::SYSCALL, Key::Arg(_, None)) => map(
             recognize(terminated(
                 many1_count(take_while1(is_hex_digit)),
                 peek(take_while1(is_sep)),
             )),
-            |s| -> Result<_, ()> {
+            |s| {
                 let ps = unsafe { str::from_utf8_unchecked(s) };
                 match u64::from_str_radix(ps, 16) {
-                    Ok(n) => Ok(PValue::Number(Number::Hex(n))),
-                    Err(_) => Ok(PValue::Str(s, Quote::None)),
+                    Ok(n) => PValue::Number(Number::Hex(n)),
+                    Err(_) => PValue::Str(s, Quote::None),
                 }
             },
         )(input)?,
@@ -312,19 +321,17 @@ fn parse_common(input: &[u8], ty: MessageType, c: Common) -> IResult<&[u8], PVal
 #[inline(always)]
 fn parse_encoded(input: &[u8]) -> IResult<&[u8], PValue> {
     alt((
-        map_res(parse_str_dq_safe, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Double))
-        }),
-        map_res(
-            terminated(
+        map(parse_str_dq_safe, |s| PValue::Str(s, Quote::Double)),
+        terminated(
+            map(
                 recognize(many1_count(take_while_m_n(2, 2, is_hex_digit))),
-                peek(take_while1(is_sep)),
+                PValue::HexStr,
             ),
-            |s| -> Result<_, ()> { Ok(PValue::HexStr(s)) },
+            peek(take_while1(is_sep)),
         ),
-        map_res(
-            terminated(alt((tag("(null)"), tag("?"))), peek(take_while1(is_sep))),
-            |_| -> Result<_, ()> { Ok(PValue::Empty) },
+        terminated(
+            value(PValue::Empty, alt((tag("(null)"), tag("?")))),
+            peek(take_while1(is_sep)),
         ),
     ))(input)
 }
@@ -346,17 +353,9 @@ fn parse_hex(input: &[u8]) -> IResult<&[u8], PValue> {
 /// Recognize decimal value
 #[inline(always)]
 fn parse_dec(input: &[u8]) -> IResult<&[u8], PValue> {
-    map_res(
-        terminated(
-            pair(opt(tag("-")), take_while1(is_digit)),
-            peek(take_while1(is_sep)),
-        ),
-        |(sign, digits)| -> Result<_, std::num::ParseIntError> {
-            let sign = if sign.is_some() { -1 } else { 1 };
-            let digits = unsafe { str::from_utf8_unchecked(digits) };
-            Ok(PValue::Number(Number::Dec(sign * i64::from_str(digits)?)))
-        },
-    )(input)
+    map(terminated(dec_i64, peek(take_while1(is_sep))), |n| {
+        PValue::Number(Number::Dec(n))
+    })(input)
 }
 
 /// Recognize octal value
@@ -379,7 +378,7 @@ fn parse_unspec_value<'a>(
 ) -> IResult<&'a [u8], PValue<'a>> {
     // work around apparent AppArmor breakage
     match (ty, name) {
-        (msg_type::SYSCALL, b"subj") | (msg_type::USER_AUTH, b"subj") => {
+        (_, b"subj") => {
             if let Ok((input, s)) = recognize(tuple((
                 opt(tag("=")),
                 parse_str_unq,
@@ -398,28 +397,18 @@ fn parse_unspec_value<'a>(
     };
 
     alt((
-        map_res(
-            terminated(take_while1(is_safe_unquoted_chr), peek(take_while1(is_sep))),
-            |s| -> Result<_, ()> { Ok(PValue::Str(s, Quote::None)) },
+        terminated(
+            map(take_while1(is_safe_unquoted_chr), |s| {
+                PValue::Str(s, Quote::None)
+            }),
+            peek(take_while1(is_sep)),
         ),
-        map_res(parse_kv_sq, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Single))
-        }),
-        map_res(parse_str_sq, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Single))
-        }),
-        map_res(parse_str_dq, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Double))
-        }),
-        map_res(parse_kv_braced, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Braces))
-        }),
-        map_res(parse_str_braced, |s| -> Result<_, ()> {
-            Ok(PValue::Str(s, Quote::Braces))
-        }),
-        map_res(peek(take_while1(is_sep)), |_| -> Result<_, ()> {
-            Ok(PValue::Empty)
-        }),
+        map(parse_kv_sq, |s| PValue::Str(s, Quote::Single)),
+        map(parse_str_sq, |s| PValue::Str(s, Quote::Single)),
+        map(parse_str_dq, |s| PValue::Str(s, Quote::Double)),
+        map(parse_kv_braced, |s| PValue::Str(s, Quote::Braces)),
+        map(parse_str_braced, |s| PValue::Str(s, Quote::Braces)),
+        value(PValue::Empty, peek(take_while1(is_sep))),
     ))(input)
 }
 
@@ -440,7 +429,7 @@ fn parse_str_dq(input: &[u8]) -> IResult<&[u8], &[u8]> {
 
 #[inline(always)]
 fn parse_str_braced(input: &[u8]) -> IResult<&[u8], &[u8]> {
-    delimited(tag("{"), take_while(|c| c != b'}'), tag("}"))(input)
+    delimited(tag("{ "), take_until(" }"), tag(" }"))(input)
 }
 
 #[inline(always)]
@@ -490,17 +479,17 @@ fn parse_kv_braced(input: &[u8]) -> IResult<&[u8], &[u8]> {
 /// Recognize regular keys of key/value pairs
 #[inline(always)]
 fn parse_key(input: &[u8]) -> IResult<&[u8], Key> {
-    map_res(
+    map(
         recognize(pair(alpha1, many0_count(alt((alphanumeric1, is_a("-_")))))),
-        |s: &[u8]| -> Result<_, ()> {
+        |s: &[u8]| {
             if let Ok(c) = Common::try_from(s) {
-                Ok(Key::Common(c))
+                Key::Common(c)
             } else if s.ends_with(b"uid") {
-                Ok(Key::NameUID(NVec::from(s)))
+                Key::NameUID(NVec::from(s))
             } else if s.ends_with(b"gid") {
-                Ok(Key::NameGID(NVec::from(s)))
+                Key::NameGID(NVec::from(s))
             } else {
-                Ok(Key::Name(NVec::from(s)))
+                Key::Name(NVec::from(s))
             }
         },
     )(input)
@@ -509,38 +498,25 @@ fn parse_key(input: &[u8]) -> IResult<&[u8], Key> {
 /// Recognize length specifier for EXECVE split arguments, e.g. a1_len
 #[inline(always)]
 fn parse_key_a_x_len(input: &[u8]) -> IResult<&[u8], Key> {
-    map_res(
-        delimited(tag("a"), digit1, tag("_len")),
-        |x| -> Result<_, std::num::ParseIntError> {
-            let x = unsafe { str::from_utf8_unchecked(x) }.parse()?;
-            Ok(Key::ArgLen(x))
-        },
-    )(input)
+    map(delimited(tag("a"), dec_u32, tag("_len")), Key::ArgLen)(input)
 }
 
 /// Recognize EXECVE split arguments, e.g. a1[3]
 #[inline(always)]
 fn parse_key_a_xy(input: &[u8]) -> IResult<&[u8], Key> {
-    map_res(
-        tuple((tag("a"), digit1, tag("["), digit1, tag("]"))),
-        |(_, x, _, y, _)| -> Result<Key, std::num::ParseIntError> {
-            let x = unsafe { str::from_utf8_unchecked(x) }.parse()?;
-            let y = unsafe { str::from_utf8_unchecked(y) }.parse()?;
-            Ok(Key::Arg(x, Some(y)))
-        },
+    map(
+        pair(
+            preceded(tag("a"), dec_u32),
+            delimited(tag("["), dec_u16, tag("]")),
+        ),
+        |(x, y)| Key::Arg(x, Some(y)),
     )(input)
 }
 
 /// Recognize SYSCALL, EXECVE regular argument keys, e.g. a1, a2, a3…
 #[inline(always)]
 fn parse_key_a_x(input: &[u8]) -> IResult<&[u8], Key> {
-    map_res(
-        preceded(tag("a"), digit1),
-        |x| -> Result<Key, std::num::ParseIntError> {
-            let x = unsafe { str::from_utf8_unchecked(x) }.parse()?;
-            Ok(Key::Arg(x, None))
-        },
-    )(input)
+    map(preceded(tag("a"), u32), |x| Key::Arg(x, None))(input)
 }
 
 /// Recognize identifiers (used in some irregular messages)
@@ -578,7 +554,7 @@ mod test {
     use super::msg_type::*;
     use super::*;
 
-    fn do_parse<T>(text: T) -> Result<(Option<Vec<u8>>, MessageType, EventID, Record), String>
+    fn do_parse<T>(text: T) -> Result<(Option<Vec<u8>>, MessageType, EventID, Record), ParseError>
     where
         T: AsRef<[u8]>,
     {
@@ -901,9 +877,13 @@ mod test {
         do_parse(include_bytes!("testdata/line-daemon-end.txt")).unwrap();
         do_parse(include_bytes!("testdata/line-netfilter.txt")).unwrap();
         do_parse(include_bytes!("testdata/line-anom-abend.txt")).unwrap();
+        do_parse(include_bytes!("testdata/line-anom-abend-2.txt")).unwrap();
         do_parse(include_bytes!("testdata/line-user-auth.txt")).unwrap();
         do_parse(include_bytes!("testdata/line-sockaddr-unix.txt")).unwrap();
+        do_parse(include_bytes!("testdata/line-sockaddr-unix-2.txt")).unwrap();
         do_parse(include_bytes!("testdata/line-user-auth-2.txt")).unwrap();
+        do_parse(include_bytes!("testdata/line-mac-policy-load.txt")).unwrap();
+        do_parse(include_bytes!("testdata/line-tty.txt")).unwrap();
     }
 
     #[test]
@@ -911,9 +891,4 @@ mod test {
     fn breakage_sockaddr_unknown() {
         do_parse(include_bytes!("testdata/line-sockaddr-unknown.txt")).unwrap();
     }
-
-    #[test]
-    fn breakage_mac_policy_load() {
-        do_parse(include_bytes!("testdata/line-mac-policy-load.txt")).unwrap();
-    }
 }
