@@ -1,12 +1,13 @@
 use std::collections::HashMap;
 use std::convert::{TryFrom, TryInto};
-use std::error::Error as StdError;
 use std::fmt::{self, Debug, Display};
 use std::iter::Iterator;
 use std::ops::Range;
 use std::str;
 use std::string::*;
 
+use indexmap::IndexMap;
+
 use lazy_static::lazy_static;
 
 use serde::ser::SerializeMap;
@@ -15,6 +16,73 @@ use serde::{Serialize, Serializer};
 use crate::constants::*;
 use crate::quoted_string::ToQuotedString;
 
+/// Collect records in [`EventBody`] context as single or multiple
+/// instances.
+///
+/// Examples for single instances are `SYSCALL`,`EXECVE` (even if the
+/// latter can be split across multiple lines). An example for
+/// multiple instances is `PATH`.
+///
+/// "Multi" records are serialized as list-of-maps (`[ { "key":
+/// "value", … }, { "key": "value", … } … ]`)
+#[derive(Debug, Clone)]
+pub enum EventValues {
+    // e.g SYSCALL, EXECVE
+    Single(Record),
+    // e.g. PATH
+    Multi(Vec<Record>),
+}
+
+impl Serialize for EventValues {
+    #[inline(always)]
+    fn serialize<S: Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
+        match self {
+            EventValues::Single(rv) => rv.serialize(s),
+            EventValues::Multi(rvs) => s.collect_seq(rvs),
+        }
+    }
+}
+
+#[derive(Clone, Debug)]
+pub struct Event {
+    pub node: Option<Vec<u8>>,
+    pub id: EventID,
+    pub body: IndexMap<MessageType, EventValues>,
+    pub filter: bool,
+}
+
+impl Event {
+    pub fn new(node: Option<Vec<u8>>, id: EventID) -> Self {
+        Event {
+            node,
+            id,
+            body: IndexMap::with_capacity(5),
+            filter: false,
+        }
+    }
+}
+
+impl Serialize for Event {
+    #[inline(always)]
+    fn serialize<S>(&self, s: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let length = self.body.len() + if self.node.is_some() { 2 } else { 1 };
+        let mut map = s.serialize_map(Some(length))?;
+        map.serialize_key("ID")?;
+        map.serialize_value(&self.id)?;
+        if let Some(node) = &self.node {
+            map.serialize_key("NODE")?;
+            map.serialize_value(&node.as_slice().to_quoted_string())?;
+        }
+        for (k, v) in &self.body {
+            map.serialize_entry(&k, &v)?;
+        }
+        map.end()
+    }
+}
+
 /// The identifier of an audit event, corresponding to the
 /// `msg=audit(…)` part of every _auditd(8)_ log line.
 ///
@@ -41,12 +109,7 @@ impl Display for EventID {
 impl Serialize for EventID {
     #[inline(always)]
     fn serialize<S: Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
-        s.collect_str(&format_args!(
-            "{}.{:03}:{}",
-            self.timestamp / 1000,
-            self.timestamp % 1000,
-            self.sequence
-        ))
+        s.collect_str(&self)
     }
 }
 
@@ -198,9 +261,9 @@ pub enum Key {
     /// translated / "enriched" values
     NameTranslated(NVec),
     /// `a0`, `a1`, `a2[0]`, `a2[1]`…
-    Arg(u16, Option<u16>),
+    Arg(u32, Option<u16>),
     /// `a0_len` …
-    ArgLen(u16),
+    ArgLen(u32),
     Literal(&'static str),
 }
 
@@ -382,7 +445,7 @@ impl Debug for Record {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let mut seq = f.debug_struct("Record");
         for (k, v) in self {
-            seq.field(&*k.to_string(), &v);
+            seq.field(&k.to_string(), &v);
         }
         seq.finish()
     }
@@ -505,7 +568,7 @@ pub struct RValue<'a> {
 }
 
 impl TryFrom<RValue<'_>> for Vec<u8> {
-    type Error = Box<dyn StdError>;
+    type Error = &'static str;
     fn try_from(v: RValue) -> Result<Self, Self::Error> {
         match v.value {
             Value::Str(r, Quote::Braces) => {
@@ -525,19 +588,17 @@ impl TryFrom<RValue<'_>> for Vec<u8> {
                 }
                 Ok(sb)
             }
-            Value::Number(_) => Err("Won't convert number to string".into()),
-            Value::List(_) | Value::StringifiedList(_) => {
-                Err("Can't convert list to scalar".into())
-            }
-            Value::Map(_) => Err("Can't convert map to scalar".into()),
-            Value::Skipped(_) => Err("Can't convert skipped to scalar".into()),
+            Value::Number(_) => Err("Won't convert number to string"),
+            Value::List(_) | Value::StringifiedList(_) => Err("Can't convert list to scalar"),
+            Value::Map(_) => Err("Can't convert map to scalar"),
+            Value::Skipped(_) => Err("Can't convert skipped to scalar"),
             Value::Literal(s) => Ok(s.to_string().into()),
         }
     }
 }
 
 impl TryFrom<RValue<'_>> for Vec<Vec<u8>> {
-    type Error = Box<dyn StdError>;
+    type Error = &'static str;
     fn try_from(value: RValue) -> Result<Self, Self::Error> {
         match value.value {
             Value::List(values) | Value::StringifiedList(values) => {
@@ -551,7 +612,7 @@ impl TryFrom<RValue<'_>> for Vec<Vec<u8>> {
                 }
                 Ok(rv)
             }
-            _ => Err("not a list".into()),
+            _ => Err("not a list"),
         }
     }
 }
